using System;
using System.Collections.Generic;
using System.DirectoryServices.Protocols;
using System.Net;
using System.Text;

namespace WorkflowActivity
{
    [Flags]
    public enum AdsUserFlags
    {
        Script = 1, // 0x1
        AccountDisabled = 2, // 0x2
        HomeDirectoryRequired = 8, // 0x8
        AccountLockedOut = 16, // 0x10
        PasswordNotRequired = 32, // 0x20
        PasswordCannotChange = 64, // 0x40
        EncryptedTextPasswordAllowed = 128, // 0x80
        TempDuplicateAccount = 256, // 0x100
        NormalAccount = 512, // 0x200
        InterDomainTrustAccount = 2048, // 0x800
        WorkstationTrustAccount = 4096, // 0x1000
        ServerTrustAccount = 8192, // 0x2000
        PasswordDoesNotExpire = 65536, // 0x10000
        MnsLogonAccount = 131072, // 0x20000
        SmartCardRequired = 262144, // 0x40000
        TrustedForDelegation = 524288, // 0x80000
        AccountNotDelegated = 1048576, // 0x100000
        UseDesKeyOnly = 2097152, // 0x200000
        DontRequirePreauth = 4194304, // 0x400000
        PasswordExpired = 8388608, // 0x800000
        TrustedToAuthenticateForDelegation = 16777216, // 0x1000000
        NoAuthDataRequired = 33554432 // 0x2000000
    }

    public class LDAP : IDisposable
    {
        #region Constructors
        public LDAP(string server, bool useCurrentCredentials)
        {
            _ldapDirectoryIdentifier = new LdapDirectoryIdentifier(server);
            if (useCurrentCredentials)
            {
                _ldapConnection = new LdapConnection(_ldapDirectoryIdentifier, System.Net.CredentialCache.DefaultNetworkCredentials, AuthType.Negotiate);
            }
            else
            {
                _ldapConnection = new LdapConnection(_ldapDirectoryIdentifier);
            }
            _ldapConnection.Bind();
        }

        public LDAP(string server, string adminUser, string adminPassword)
        {
            _ldapDirectoryIdentifier = new LdapDirectoryIdentifier(server);
            NetworkCredential cred = null;
            if (adminUser.IndexOf('\\') == -1)
            {
                cred = new NetworkCredential(adminUser, adminPassword);
            }
            else
            {
                cred = new NetworkCredential(adminUser.Substring(adminUser.IndexOf('\\') + 1), adminPassword, adminUser.Substring(0, adminUser.IndexOf('\\')));
            }
            _ldapConnection = new LdapConnection(_ldapDirectoryIdentifier, cred, AuthType.Negotiate);
            _ldapConnection.Bind();
        }
        #endregion

        #region Constants
        public const string LDAP_ATTR_OBJECT_CLASS = "objectClass";
        public const string LDAP_ATTR_ACCOUNT_NAME = "sAMAccountName";
        public const string LDAP_ATTR_PASSWORD = "unicodePwd";
        public const string LDAP_ATTR_USER_PRINCIPAL_NAME = "userPrincipalName";
        public const string LDAP_ATTR_USER_ACCOUNT_CONTROL = "userAccountControl";
        public const string LDAP_ATTR_DISTINGUISHED_NAME = "distinguishedName";
        public const string LDAP_ATTR_EXCH_HOME_MDB = "homeMDB";
        public const string LDAP_ATTR_EXCH_HOME_MTA = "homeMTA";
        public const string LDAP_ATTR_MAIL = "mail";
        public const string LDAP_ATTR_EXCH_MAIL_NICKNAME = "mailNickname";
        public const string LDAP_ATTR_EXCH_PROXY_ADDR = "proxyAddresses";
        public const string LDAP_ATTR_DISPLAY_NAME = "displayName";
        public const string LDAP_ATTR_EMPLOYEE_ID = "employeeID";
        public const string LDAP_ATTR_MANAGER = "manager"; // Distinguished name
        public const string LDAP_ATTR_DIALIN = "msNPAllowDialin";
        public const string LDAP_ATTR_HOME_DIRECTORY = "homeDirectory";
        public const string LDAP_ATTR_PROFILE_PATH = "profilePath";
        public const string LDAP_ATTR_LOGIN_SCRIPT = "scriptPath";
        public const string LDAP_ATTR_OFC_COUNTRY = "co";
        public const string LDAP_ATTR_OFC_COUNTRY_CODE = "countryCode";
        public const string LDAP_ATTR_OFC_CITY = "l";
        public const string LDAP_ATTR_OFC_POSTALCODE = "postalCode";
        public const string LDAP_ATTR_OFC_POBOX = "postOfficeBox";
        public const string LDAP_ATTR_OFC_STATE = "st";
        public const string LDAP_ATTR_OFC_STREET_ADDR = "streetAddress";
        public const string LDAP_ATTR_OFC_PHONE = "telephoneNumber";
        public const string LDAP_ATTR_FIRST_NAME = "givenName";
        public const string LDAP_ATTR_LAST_NAME = "sn";
        public const string LDAP_ATTR_ACCT_EXPIRES = "accountExpires";
        public const string LDAP_ATTR_MEMBERS = "member";

        public const string LDAP_VALUE_USERCLASS = "user";
        public const string LDAP_VALUE_GROUPCLASS = "group";

        #endregion
        protected LdapDirectoryIdentifier _ldapDirectoryIdentifier = null;
        protected LdapConnection _ldapConnection = null;

        public bool CreateUser(string ldapRootPath, string ldapPath, string name, string password, string email, AdsUserFlags flags, string[] groups, Dictionary<string, string> attributes)
        {
            ldapRootPath = EnsureDistinguished(ldapRootPath);
            ldapPath = EnsureDistinguished(ldapPath);

            if (FindUser(ldapPath, name) != null) throw new ApplicationException("User Already Exists");
            string userDistinguished = CreateUserObjectInAD(ldapPath, name, password, email, flags, attributes);
            if (userDistinguished == null) return false;
            if (groups != null)
            {
                return (AddUserToGroups(userDistinguished, ldapRootPath, groups));
            }
            else
            {
                return true;
            }
        }

        public string FindUser(string ldapPath, string name)
        {
            return (FindUser(ldapPath, name, true));
        }

        public string FindUser(string ldapPath, string name, bool includeSub)
        {
            ldapPath = EnsureDistinguished(ldapPath);

            System.DirectoryServices.Protocols.SearchScope scope = System.DirectoryServices.Protocols.SearchScope.Base;
            if (includeSub) scope = System.DirectoryServices.Protocols.SearchScope.Subtree;

            string filter = string.Format("(&(objectClass={0})(sAMAccountName={1}))", LDAP_VALUE_USERCLASS, name);

            SearchRequest req = new SearchRequest(ldapPath, filter, scope, "sAMAccountName");

            SearchResponse resp = (SearchResponse)_ldapConnection.SendRequest(req);

            if (resp.Entries.Count == 1) return resp.Entries[0].DistinguishedName;
            else return null;

        }

        public string CreateUserObjectInAD(string path, string name, string password, string email, AdsUserFlags flags, Dictionary<string, string> attributes)
        {
            path = EnsureDistinguished(path);
            string fullyQualifiedName = "cn=" + name + "," + path;
            List<DirectoryAttribute> lstDA = new List<DirectoryAttribute>();
            lstDA.Add(new DirectoryAttribute(LDAP_ATTR_OBJECT_CLASS, LDAP_VALUE_USERCLASS));
            lstDA.Add(new DirectoryAttribute(LDAP_ATTR_ACCOUNT_NAME, name));
            lstDA.Add(new DirectoryAttribute(LDAP_ATTR_USER_PRINCIPAL_NAME, email));
            lstDA.Add(new DirectoryAttribute(LDAP_ATTR_USER_ACCOUNT_CONTROL, Convert.ToString((int)flags)));
            lstDA.Add(new DirectoryAttribute(LDAP_ATTR_PASSWORD, GetPasswordData(password)));
            if (attributes != null)
            {
                foreach (KeyValuePair<string, string> kvp in attributes)
                {
                    lstDA.Add(new DirectoryAttribute(kvp.Key, kvp.Value));
                }
            }
            try
            {
                AddRequest req = new AddRequest(fullyQualifiedName, lstDA.ToArray());

                AddResponse resp = (AddResponse)_ldapConnection.SendRequest(req);

                if (resp.ResultCode != ResultCode.Success)
                {
                    return null;
                }
                else
                {
                    return fullyQualifiedName;
                }

            }
            catch (Exception e)
            {
                throw e;
            }

        }

        public bool AddUserToGroups(string userDistinguishedName, string ldapRoot, string[] groupNames)
        {
            ldapRoot = EnsureDistinguished(ldapRoot);
            bool result = true;
            for (int looper = 0; looper < groupNames.Length; looper++)
            {
                string groupDistinguished = FindGroup(ldapRoot, groupNames[looper]);
                if (groupDistinguished == null) throw new ApplicationException("Group " + groupNames[looper] + " was not found.");
                result &= AddUserToGroup(userDistinguishedName, groupDistinguished);
            }
            return result;
        }

        public string FindGroup(string ldapPath, string name)
        {
            return (FindGroup(ldapPath, name, true));
        }

        public string FindGroup(string ldapPath, string name, bool includeSub)
        {
            ldapPath = EnsureDistinguished(ldapPath);

            System.DirectoryServices.Protocols.SearchScope scope = System.DirectoryServices.Protocols.SearchScope.Base;
            if (includeSub) scope = System.DirectoryServices.Protocols.SearchScope.Subtree;
            string filter = string.Format("(&(objectClass={0})(sAMAccountName={1}))", LDAP_VALUE_GROUPCLASS, name);

            SearchRequest req = new SearchRequest(ldapPath, filter, scope, "sAMAccountName");

            SearchResponse resp = (SearchResponse)_ldapConnection.SendRequest(req);

            if (resp.Entries.Count == 1) return resp.Entries[0].DistinguishedName;
            else return null;
        }

        public bool AddUserToGroup(string userDistinguished, string groupDistinguished)
        {
            // Modify Group
            ModifyRequest req = new ModifyRequest(groupDistinguished, DirectoryAttributeOperation.Add, LDAP_ATTR_MEMBERS, new string[] { userDistinguished });
            ModifyResponse resp = (ModifyResponse)_ldapConnection.SendRequest(req);

            if (resp.ResultCode == ResultCode.Success) return true;
            else return false;
        }

        public bool EnableMailbox(string userDistinguished, string[] emailAddresses)
        {
            // TODO
            return true;
        }

        protected string EnsureDistinguished(string ldapPath)
        {
            if (ldapPath.IndexOf("LDAP://") >= 0) return (ldapPath.Substring("LDAP://".Length));
            if (ldapPath.IndexOf("LDAPS://") >= 0) return (ldapPath.Substring("LDAPS://".Length));
            return (ldapPath);
        }

        #region Code from DirectoryServices.Net Listing 10.16
        private static void ChangePassword(DirectoryConnection connection, string userDN, string oldPassword, string newPassword)
        {
            DirectoryAttributeModification deleteMod = new DirectoryAttributeModification();
            deleteMod.Name = "unicodePwd";
            deleteMod.Add(GetPasswordData(oldPassword));
            deleteMod.Operation = DirectoryAttributeOperation.Delete;
            DirectoryAttributeModification addMod =
            new DirectoryAttributeModification();
            addMod.Name = "unicodePwd";
            addMod.Add(GetPasswordData(newPassword));
            addMod.Operation = DirectoryAttributeOperation.Add;
            ModifyRequest request = new ModifyRequest(userDN, deleteMod, addMod);
            DirectoryResponse response = connection.SendRequest(request);
        }
        private static void SetPassword(DirectoryConnection connection, string userDN, string password)
        {
            DirectoryAttributeModification pwdMod =
            new DirectoryAttributeModification();
            pwdMod.Name = "unicodePwd";
            pwdMod.Add(GetPasswordData(password));
            pwdMod.Operation = DirectoryAttributeOperation.Replace;
            ModifyRequest request = new ModifyRequest(userDN, pwdMod);
            DirectoryResponse response = connection.SendRequest(request);
        }

        private static byte[] GetPasswordData(string password)
        {
            string formattedPassword;
            formattedPassword = String.Format("\"{0}\"", password);
            return (Encoding.Unicode.GetBytes(formattedPassword));
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            if (_ldapConnection != null) _ldapConnection.Dispose();
        }

        #endregion
    }
}
